#include "SPI.h"
#include "TFT_eSPI.h" // Graphics and font library for ILI9341 driver chip
#include "Free_Fonts.h" // Include the header file attached to this sketch
#include <EmonLib.h> // Include EmonLib Library

#include <WiFi.h>
#include <ESPAsyncWebServer.h>
#include <AsyncTCP.h>

AsyncWebServer server(80);
AsyncWebSocket ws("/ws");

EnergyMonitor emon1; // Create an instance for energy monitoring

// Use hardware SPI
TFT_eSPI tft = TFT_eSPI(); 

TFT_eSprite sprite = TFT_eSprite(&tft); // Create sprite object "sprite"

int currentX = 30; // Start position adjusted to the left
int graphStartX = 30; // Start position of the graph adjusted to the left
int numberStartX = 15; // Start position of the numbers adjusted to the left

// Variables for energy calculation
double realPower = 0; // Variable to store real power (Watts)
double supplyVoltage = 230.0; // Assuming a 230V supply voltage (adjust according to your region)
double current = 0; // Variable to store current (mA)
double energy = 0; // Variable to store total energy usage (kWh)
unsigned long lastUpdateTime = 0;
unsigned long updateInterval = 1000; // Update interval in milliseconds
double unitRate = 0; // Unit rate (cost per kWh)
double totalBill = 0; // Variable to store total bill

int maxCurrent = 20; // Initial max current value for scaling
int maxPower = 100; // Initial max power value for scaling

const char index_html[] PROGMEM = R"rawliteral(
<!DOCTYPE HTML>
<html>
<head>
  <title>Energy Monitor Dashboard</title>
  <style>
    body {
      font-family: Arial, sans-serif;
      text-align: center;
      background-color: #f0f0f0;
    }
    .card {
      background-color: white;
      box-shadow: 0 4px 8px rgba(0, 0, 0, 0.1);
      border-radius: 10px;
      padding: 20px;
      margin: 20px auto;
      width: 300px;
      text-align: center;
    }
    .card h2 {
      font-size: 24px;
    }
    .card p {
      font-size: 18px;
      font-weight: bold;
    }
    #rateForm {
      margin-bottom: 20px;
    }
    #rateForm input {
      padding: 10px;
      border-radius: 5px;
      border: 1px solid #ddd;
      font-size: 16px;
    }
    #rateForm button {
      padding: 10px 20px;
      border-radius: 5px;
      border: none;
      background-color: #28a745;
      color: white;
      font-size: 16px;
      cursor: pointer;
    }
  </style>
</head>
<body>
  <div id="rateForm">
    <input type="number" id="unitRate" placeholder="Enter unit rate">
    <button onclick="setRate()">Set Rate</button>
  </div>
  <div class="card">
    <h2>Current</h2>
    <p id="current">0 mA</p>
  </div>
  <div class="card">
    <h2>Real Power</h2>
    <p id="realPower">0 W</p>
  </div>
  <div class="card">
    <h2>Energy</h2>
    <p id="energy">0 kWh</p>
  </div>
  <div class="card">
    <h2>Total Bill</h2>
    <p id="totalBill">0 Rs</p>
  </div>

  <script>
    var gateway = `ws://${window.location.hostname}/ws`;
    var websocket;

    window.addEventListener('load', onLoad);

    function onLoad(event) {
      initWebSocket();
    }

    function initWebSocket() {
      websocket = new WebSocket(gateway);
      websocket.onopen = onOpen;
      websocket.onclose = onClose;
      websocket.onmessage = onMessage;
    }

    function onOpen(event) {
      console.log('Connection opened');
    }

    function onClose(event) {
      console.log('Connection closed');
      setTimeout(initWebSocket, 2000);
    }

    function onMessage(event) {
      var data = JSON.parse(event.data);
      document.getElementById('current').innerText = data.current + ' mA';
      document.getElementById('realPower').innerText = data.realPower + ' W';
      document.getElementById('energy').innerText = data.energy + ' kWh';
      document.getElementById('totalBill').innerText = data.totalBill + ' Rs';
    }

    function setRate() {
      var unitRate = document.getElementById('unitRate').value;
      websocket.send('setRate:' + unitRate);
    }
  </script>
</body>
</html>
)rawliteral";

void notifyClients() {
  String json = "{\"current\":" + String(current) + ",\"realPower\":" + String(realPower) + ",\"energy\":" + String(energy) + ",\"totalBill\":" + String(totalBill) + "}";
  ws.textAll(json);
}

void handleWebSocketMessage(void *arg, uint8_t *data, size_t len) {
  AwsFrameInfo *info = (AwsFrameInfo*)arg;
  if(info->final && info->index == 0 && info->len == len && info->opcode == WS_TEXT) {
    data[len] = 0;
    String message = (char*)data;
    if(message.startsWith("setRate:")) {
      unitRate = message.substring(8).toDouble();
    }
  }
}

void onEvent(AsyncWebSocket *server, AsyncWebSocketClient *client, AwsEventType type, void *arg, uint8_t *data, size_t len) {
  if(type == WS_EVT_DATA) {
    handleWebSocketMessage(arg, data, len);
  }
}

void setup(void) {
  Serial.begin(115200);
  tft.begin();
  tft.setRotation(1);
  tft.fillScreen(TFT_BLACK); // Clear the screen

  sprite.createSprite(240, 120); // Create a sprite of 240x120 pixels
  sprite.setTextColor(TFT_WHITE, TFT_BLACK); // Set text color to white with black background
  sprite.setFreeFont(&FreeSans9pt7b); // Use FreeSans9pt7b font

  emon1.current(11, 60.6); // Current: input pin 11, calibration for SCT-013

  Serial.println("Enter the unit rate (cost per kWh): ");
  while (Serial.available() == 0) {
    // Wait for user input
  }
  unitRate = Serial.parseFloat();
  Serial.print("Unit rate set to: ");
  Serial.print(unitRate);
  Serial.println(" per kWh");

  // Configure the ESP32 as an access point
  WiFi.softAP("IOTmeter");
  IPAddress IP = WiFi.softAPIP();
  Serial.print("AP IP address: ");
  Serial.println(IP);

  ws.onEvent(onEvent);
  server.addHandler(&ws);

  // Serve the HTML page
  server.on("/", HTTP_GET, [](AsyncWebServerRequest *request){
    request->send_P(200, "text/html", index_html);
  });

  server.begin();

  drawGraphBase();
}

void loop() {
  // Calculate real power and energy usage
  current = emon1.calcIrms(1480) * 1000; // Calculate current in mA
  realPower = current * supplyVoltage / 1000; // Calculate real power in Watts

  // Calculate energy usage
  unsigned long currentTime = millis();
  if (currentTime - lastUpdateTime >= updateInterval) {
    // Energy (kWh) = Power (W) * Time (h)
    energy += (realPower * (updateInterval / 1000.0 / 3600.0));
    lastUpdateTime = currentTime;

    // Calculate total bill
    totalBill = energy * unitRate;

    // Clear sprite and set text properties
    sprite.fillSprite(TFT_BLACK); // Clear the sprite with black color
    sprite.setCursor(5, 15); // Set cursor to top-left of the sprite

    // Draw values
    sprite.println("Current: " + String(current, 2) + " mA");
    sprite.println("Power: " + String(realPower, 2) + " W");
    sprite.println("Total: " + String(energy, 2) + " kWh");
    sprite.println("Bill: " + String(totalBill, 2) + " Rs");

    // Push the sprite to the bottom half of the screen
    sprite.pushSprite(graphStartX, 132); // Adjusted position to avoid overlap
    
    // Notify clients
    notifyClients();
  }

  // Adjust graph scaling if needed
  if (current > maxCurrent) {
    maxCurrent = current + 5; // Add some margin
    drawGraphBase();
  }
  if (realPower > maxPower) {
    maxPower = realPower + 10; // Add some margin
    drawGraphBase();
  }

  // Plot the graph on the top half of the display
  tft.drawLine(currentX, 120 - (current * 100 / maxCurrent), currentX + 1, 120 - (emon1.calcIrms(1480) * 1000 * 100 / maxCurrent), TFT_RED); // Draw current in red
  tft.drawLine(currentX, 120 - (realPower * 100 / maxPower) / 2, currentX + 1, 120 - ((current * supplyVoltage / 1000) * 100 / maxPower) / 2, TFT_BLUE); // Draw power in blue

  currentX += 1;
  if (currentX >= tft.width()) {
    currentX = 30; // Reset to start position
    tft.fillRect(graphStartX, 0, 240 - graphStartX, 120, TFT_BLACK); // Clear the top half of the screen with black, adjusted starting position

    drawGraphBase();
  }

  // Small delay to ensure smooth updating of the graph
  delay(100); // 100ms delay for graph plotting
}

void drawGraphBase() {
  // Clear the area for the graph base
  tft.fillRect(0, 0, 240, 120, TFT_BLACK);

  // Draw green line as background
  tft.drawFastHLine(30, 120 - 40, tft.width() - 30, TFT_GREEN);

  // Draw current scale
  for (int i = 0; i <= maxCurrent; i += maxCurrent / 5) {
    tft.setCursor(numberStartX, 120 - (i * 100 / maxCurrent) - 3); // Position the cursor above the green line
    tft.setTextColor(TFT_WHITE);
    tft.print(i);
  }

  // Draw power scale on the other side
  for (int i = 0; i <= maxPower; i += maxPower / 5) {
    tft.setCursor(tft.width() - numberStartX - 20, 120 - (i * 100 / maxPower) - 3); // Position the cursor above the green line
    tft.setTextColor(TFT_WHITE);
    tft.print(i);
  }
}
