#include <WiFi.h>
#include <ESPAsyncWebServer.h>
#include <AsyncTCP.h>
#include <EmonLib.h>                   // Include EmonLib Library

const char* ssid = "JioFiber-4G";
const char* password = "1234567890";

AsyncWebServer server(80);
AsyncWebSocket ws("/ws");

EnergyMonitor emon1;                   // Create an instance for energy monitoring

// Variables for energy calculation
double realPower = 0;                  // Variable to store real power (Watts)
double supplyVoltage = 230.0;          // Assuming a 230V supply voltage (adjust according to your region)
double current = 0;                    // Variable to store current (mA)
double energy = 0;                     // Variable to store total energy usage (kWh)
unsigned long lastUpdateTime = 0;
unsigned long updateInterval = 1000;   // Update interval in milliseconds
double unitRate = 0;                   // Unit rate (cost per kWh)
double totalBill = 0;                  // Variable to store total bill

const char index_html[] PROGMEM = R"rawliteral(
<!DOCTYPE HTML>
<html>
<head>
  <title>Energy Monitor Dashboard</title>
  <style>
    body {
      font-family: Arial, sans-serif;
      text-align: center;
      background-color: #f0f0f0;
    }
    .card {
      background-color: white;
      box-shadow: 0 4px 8px rgba(0, 0, 0, 0.1);
      border-radius: 10px;
      padding: 20px;
      margin: 20px auto;
      width: 300px;
      text-align: center;
    }
    .card h2 {
      font-size: 24px;
    }
    .card p {
      font-size: 18px;
      font-weight: bold;
    }
    #rateForm {
      margin-bottom: 20px;
    }
    #rateForm input {
      padding: 10px;
      border-radius: 5px;
      border: 1px solid #ddd;
      font-size: 16px;
    }
    #rateForm button {
      padding: 10px 20px;
      border-radius: 5px;
      border: none;
      background-color: #28a745;
      color: white;
      font-size: 16px;
      cursor: pointer;
    }
  </style>
</head>
<body>
  <div id="rateForm">
    <input type="number" id="unitRate" placeholder="Enter unit rate">
    <button onclick="setRate()">Set Rate</button>
  </div>
  <div class="card">
    <h2>Current</h2>
    <p id="current">0 mA</p>
  </div>
  <div class="card">
    <h2>Real Power</h2>
    <p id="realPower">0 W</p>
  </div>
  <div class="card">
    <h2>Energy</h2>
    <p id="energy">0 kWh</p>
  </div>
  <div class="card">
    <h2>Total Bill</h2>
    <p id="totalBill">0 Rs</p>
  </div>

  <script>
    var gateway = `ws://${window.location.hostname}/ws`;
    var websocket;

    window.addEventListener('load', onLoad);

    function onLoad(event) {
      initWebSocket();
    }

    function initWebSocket() {
      websocket = new WebSocket(gateway);
      websocket.onopen = onOpen;
      websocket.onclose = onClose;
      websocket.onmessage = onMessage;
    }

    function onOpen(event) {
      console.log('Connection opened');
    }

    function onClose(event) {
      console.log('Connection closed');
      setTimeout(initWebSocket, 2000);
    }

    function onMessage(event) {
      var data = JSON.parse(event.data);
      document.getElementById('current').innerText = data.current + ' mA';
      document.getElementById('realPower').innerText = data.realPower + ' W';
      document.getElementById('energy').innerText = data.energy + ' kWh';
      document.getElementById('totalBill').innerText = data.totalBill + ' Rs';
    }

    function setRate() {
      var unitRate = document.getElementById('unitRate').value;
      websocket.send('setRate:' + unitRate);
    }
  </script>
</body>
</html>
)rawliteral";

void notifyClients() {
  String json = "{\"current\":" + String(current) + ",\"realPower\":" + String(realPower) + ",\"energy\":" + String(energy) + ",\"totalBill\":" + String(totalBill) + "}";
  ws.textAll(json);
}

void handleWebSocketMessage(void *arg, uint8_t *data, size_t len) {
  AwsFrameInfo *info = (AwsFrameInfo*)arg;
  if(info->final && info->index == 0 && info->len == len && info->opcode == WS_TEXT) {
    data[len] = 0;
    String message = (char*)data;
    if(message.startsWith("setRate:")) {
      unitRate = message.substring(8).toDouble();
    }
  }
}

void onEvent(AsyncWebSocket *server, AsyncWebSocketClient *client, AwsEventType type, void *arg, uint8_t *data, size_t len) {
  if(type == WS_EVT_DATA) {
    handleWebSocketMessage(arg, data, len);
  }
}

void setup() {
  Serial.begin(115200);

  // Connect to Wi-Fi
  WiFi.begin(ssid, password);
  while (WiFi.status() != WL_CONNECTED) {
    delay(1000);
    Serial.println("Connecting to WiFi...");
  }
  Serial.println("Connected to WiFi");
  Serial.println(WiFi.localIP());


  // Initialize the SCT-013 current sensor
  emon1.current(11, 60.6);             // Current: input pin 10, calibration for SCT-013

  // Initialize WebSocket
  ws.onEvent(onEvent);
  server.addHandler(&ws);

  // Serve the HTML page
  server.on("/", HTTP_GET, [](AsyncWebServerRequest *request){
    request->send_P(200, "text/html", index_html);
  });

  server.begin();
}

void loop() {
  // Calculate real power and energy usage
  current = random(800,1000);//emon1.calcIrms(1480) * 1000;  // Calculate current in mA
  realPower = current * supplyVoltage / 1000; // Calculate real power in Watts

  // Calculate energy usage
  unsigned long currentTime = millis();
  if (currentTime - lastUpdateTime >= updateInterval) {
    // Energy (kWh) = Power (W) * Time (h)
    energy += (realPower * (updateInterval / 1000.0 / 3600.0));
    lastUpdateTime = currentTime;

    // Calculate total bill
    totalBill = energy * unitRate;

    // Notify clients
    notifyClients();
  }
  delay(1000);  // Delay for a second before the next reading
}
